NO_PLAN = 251
INVALID_PLAN = 252
INTERRUPTED = 253
UNKNOWN_ERROR = 254


UNKNOWN_ERROR_MSG = (
    'Caught unexpected error, see traceback and description below. '
    'If this seems like a bug to you, please consider creating a issue in '
    f'https://github.com/kangasta/yaml_requests/issues.\n'
)


class YamlRequestsError(Exception):
    def __init__(self, message, exit_code):
        super().__init__(message)
        self.exit_code = exit_code


class NoPlanError(YamlRequestsError):
    def __init__(self, path=None):
        if not path:
            super().__init__('No requests plan file provided.', NO_PLAN)
        else:
            super().__init__(f'Did not find plan file in {path}.', NO_PLAN)


class InvalidPlanError(YamlRequestsError):
    def __init__(self, message):
        super().__init__(message, INVALID_PLAN)


class InterruptedError(YamlRequestsError):
    def __init__(self):
        super().__init__('', INTERRUPTED)



from argparse import ArgumentParser
import json
import yaml


def get_argparser():
    parser = ArgumentParser()
    parser.add_argument(
        'plan_file',
        type=str,
        nargs='?',
        help='Load requests plan from JSON or YAML file.')
    parser.add_argument(
        '-v', '--variable',
        action='append',
        dest='variables',
        metavar='NAME:VALUE',
        help=(
            'Define a variable to be used in the requests. '
            'To set multiple variables, use this argument multiple times'))
    parser.add_argument(
        '--no-animation',
        dest='animation',
        action='store_false',
        help='Disable progress animations in console output.')
    parser.add_argument(
        '--no-colors',
        dest='colors',
        action='store_false',
        help='Disable colors in console output.')
    parser.add_argument(
        '--version',
        action='store_true',
        help='Print version information.')

    return parser


def load_plan_file(filename):
    if not filename:
        raise ValueError('No input file given.')

    with open(filename, 'r') as f:
        if filename.endswith('.json'):
            plan = json.load(f)
        elif filename.endswith('.yaml') or filename.endswith('.yml'):
            plan = yaml.load(f, Loader=yaml.SafeLoader)
        else:
            raise ValueError(
                'Failed to recognize file type. '
                'File extension must be json, yaml, or yml.')

    return plan


def parse_variables(raw_variables):
    variables = {}

    if not raw_variables:
        return variables

    for raw_variable in raw_variables:
        try:
            key, value = raw_variable.split(':', maxsplit=1)
            variables[key] = value
        except ValueError:
            raise ValueError(
                f'Variable definition "{raw_variable}" has invalid format. '
                'Variables should be defined as NAME:VALUE strings.')

    return variables



from os import system
import platform
from traceback import print_exc

from jinja2 import __version__ as _jinja2_version
from yaml import __version__ as _pyyaml_version
from requests import __version__ as _requests_version

from . import __version__
from .utils.args import get_argparser, load_plan_file, parse_variables
from .logger import ConsoleLogger
from ._plan import Plan
from ._runner import PlanRunner
from .error import (
    NoPlanError,
    InterruptedError,
    InvalidPlanError,
    YamlRequestsError,
    INVALID_PLAN,
    UNKNOWN_ERROR,
    UNKNOWN_ERROR_MSG,
)


def _print_versions():
    print(
        f'{__version__} ('
        f'jinja2={_jinja2_version}',
        f'pyyaml={_pyyaml_version}',
        f'requests={_requests_version}'
        ')')


def main():
    '''Run the application.

    Returns:
        int: Exit code
    '''

    # On windows, run shell command to enable ANSI code support
    if platform.system() == 'Windows':
        system('')

    args = get_argparser().parse_args()

    if args.version:
        _print_versions()
        return 0

    logger = ConsoleLogger(animations=args.animation, colors=args.colors)

    try:
        variables_override = parse_variables(args.variables)
    except ValueError as error:
        logger.error(str(error))
        return INVALID_PLAN

    try:
        num_errors = run(args.plan_file, logger, variables_override)
        return min(num_errors, 250)
    except YamlRequestsError as error:
        logger.error(str(error))
        return error.exit_code
    except BaseException:
        logger.close()
        logger.error(UNKNOWN_ERROR_MSG)
        print_exc()
        return UNKNOWN_ERROR


def execute():
    '''Run the application and exit with suitable exit code.

    This is the entrypoint of the application.
    '''
    code = main()
    exit(code)


def run(plan_file, logger, variables_override=None):
    try:
        if not plan_file:
            raise NoPlanError()

        try:
            plan_dict = load_plan_file(plan_file)
            plan = Plan(plan_dict, variables_override=variables_override)
        except FileNotFoundError:
            raise NoPlanError(plan_file)
        except (ValueError, AssertionError,) as error:
            raise InvalidPlanError(str(error))

        runner = PlanRunner(plan, logger)
        return runner.run()
    except KeyboardInterrupt:
        logger.close()
        raise InterruptedError()



import json

from jinja2 import Environment as _J2_Environment, StrictUndefined


def to_json_filter(value):
    str(value)  # Raises UndefinedError if value is StrictUndefined
    return json.dumps(value)


class Environment(_J2_Environment):
    def __init__(self, *args, **kwargs):
        kwargs = {
            'undefined': StrictUndefined,
            **kwargs,
        }

        super().__init__(*args, **kwargs)

        self.filters['to_json'] = to_json_filter

    def register(self, name, value):
        self.globals[name] = value

    def get(self, name):
        return self.globals.get(name)

    def _contains_template(self, str_in):
        has_start = self.variable_start_string in str_in
        has_end = self.variable_end_string in str_in
        return has_start and has_end

    def _is_template(self, str_in):
        start_eq = str_in.startswith(self.variable_start_string)
        end_eq = str_in.endswith(self.variable_end_string)
        single_start_eq = str_in.count(self.variable_start_string) == 1
        return start_eq and single_start_eq and end_eq

    def _resolve_string(self, str_in) -> any:
        if not self._contains_template(str_in):
            return str_in

        if self._is_template(str_in):
            # If input is a template, append to_json filter to maintain
            # original data type. This requires wrapping template expression
            # from user in parentheses to avoid issues with operator
            # predendence.
            str_in = str_in.replace('{{', '{{ (')
            str_in = str_in.replace('}}', ') | to_json }}')

        template = self.from_string(str_in)
        rendered = template.render()

        if 'to_json' in str_in:
            try:
                return json.loads(rendered)
            except Exception:
                pass

        return rendered

    def _resolve_dict(self, item) -> dict:
        return {key: self.resolve_templates(value)
                for key, value in item.items()}

    def _resolve_list(self, item) -> list:
        return [self.resolve_templates(i) for i in item]

    def resolve_templates(self, item) -> any:
        if isinstance(item, str):
            return self._resolve_string(item)
        elif isinstance(item, list):
            return self._resolve_list(item)
        elif isinstance(item, dict):
            return self._resolve_dict(item)
        else:
            return item

    def resolve_expression(self, expr):
        return self.compile_expression(expr)()



class RequestLogger:
    def __init__(self):
        self._requests = []

    @property
    def requests(self):
        return self._requests

    def error(self, error):
        raise error

    def title(self, name, num_requests, repeat_index=None):
        pass

    def start_request(self, request):
        self._requests.append(request)

    def close(self):
        pass

    def finish_request(self, request):
        self._requests[-1] = request



from requests import request, Session
from requests.cookies import cookiejar_from_dict

from .utils.template import Environment
from ._request import Request


class PlanRunner:
    def __init__(self, plan, logger):
        self._plan = plan
        self._env = Environment()
        self._prepare_session()

        for name, value in self._env.resolve_templates(
                self._plan.variables).items():
            self._env.register(name, value)

        self._logger = logger

    def _prepare_session(self):
        session_option = self._plan.options.session
        if not session_option:
            self._session = None
            return

        self._session = Session()

        if isinstance(session_option, dict):
            session_dict = self._env.resolve_templates(session_option)
            self._session.headers.update(session_dict.get('headers', {}))
            self._session.cookies.update(
                cookiejar_from_dict(
                    session_dict.get(
                        'cookies', {})))

    def _request(self, *args, **kwargs):
        if self._plan.options.session:
            return self._session.request(*args, **kwargs)
        return request(*args, **kwargs)

    def _has_repeat_condition(self):
        return bool(self._plan.options.repeat_while)

    def _check_repeat_condition(self):
        repeat_while = self._plan.options.repeat_while
        if isinstance(repeat_while, str):
            return bool(self._env.resolve_expression(repeat_while))

        return bool(repeat_while)

    def run(self):
        num_errors = 0
        repeat_index = 0 if self._has_repeat_condition() else None

        break_repeat = False
        repeat_while = True

        ignore_errors = self._plan.options.ignore_errors

        while repeat_while and not break_repeat:
            self._env.register('repeat_index', repeat_index)
            self._logger.title(
                self._plan.name,
                len(self._plan.requests),
                repeat_index=repeat_index)

            for request_dict in self._plan.requests:
                skip = not ignore_errors and num_errors > 0

                request = Request(request_dict, self._env, skip)

                if request.state is None:
                    self._logger.start_request(request)
                    request.send(self._request)

                self._logger.finish_request(request)

                if not request.state.ok:
                    num_errors += 1

            break_repeat = not ignore_errors and num_errors > 0
            repeat_while = (
                True if repeat_index == 0 else self._check_repeat_condition())

            if self._has_repeat_condition():
                repeat_index += 1

        return num_errors
